@@ -1,7 +1,7 @@
 /*
  * Copyright 2000  ATI Technologies Inc., Markham, Ontario,
  * Copyright 2000  VA Linux Systems Inc., Fremont, California.
- * Copyright 2007  Luc Verhaegen <lverhaegen@novell.com>
+ * Copyright 2007  Luc Verhaegen <libv@exsuse.de>
  * Copyright 2007  Matthias Hopf <mhopf@novell.com>
  * Copyright 2007  Egbert Eich   <eich@novell.com>
  * Copyright 2007  Advanced Micro Devices, Inc.
@@ -57,6 +57,11 @@
 /* GLX/DRI/DRM definitions */
 #define _XF86DRI_SERVER_
 #include "dri.h"
+/* Workaround for header mismatches */
+#ifndef DEPRECATED
+#  define DEPRECATED __attribute__ ((deprecated))
+#  define __user
+#endif
 #include "radeon_drm.h"
 #include "GL/glxint.h"
 #include "GL/glxtokens.h"
@@ -76,12 +81,14 @@
 #include "rhd_dri.h"
 #include "r5xx_accel.h"
 #include "r5xx_regs.h"
+#include "r6xx_accel.h"
 #include "rhd_cs.h"
 
 #define IS_RADEONHD_DRIVER 1
 #include "radeon_dri.h"
 
-#ifdef RANDR_12_SUPPORT		// FIXME check / move to rhd_randr.c
+#ifdef RANDR_12_SUPPORT		/* FIXME check / move to rhd_randr.c */
+# include "xf86i2c.h" /* this is complete BS, stop using unnamed structs! */
 # include "xf86Crtc.h"
 #endif
 
@@ -115,7 +122,6 @@ typedef struct {
     int dummy;
 } RADEONDRIContextRec, *RADEONDRIContextPtr;
 
-
 /* driver data only needed by dri */
 struct rhdDri {
     int               scrnIndex;
@@ -152,7 +158,7 @@ struct rhdDri {
     drmAddress        ring;             /* Map */
     int               ringSizeLog2QW;
 
-    // TODO: what is r/o ring space for (1 page)
+    /* TODO: what is r/o ring space for (1 page) */
     unsigned long     ringReadOffset;   /* Offset into GART space */
     drm_handle_t      ringReadPtrHandle; /* Handle from drmAddMap */
     drmSize           ringReadMapSize;  /* Size of map */
@@ -380,23 +386,35 @@ static void RHDEnterServer(ScreenPtr pScreen)
     if (rhdPtr->EXAInfo)
 	exaMarkSync(pScrn->pScreen);
 #endif
-    if (rhdPtr->XAAInfo)
-	SET_SYNC_FLAG(rhdPtr->XAAInfo);
 
     pSAREAPriv = (drm_radeon_sarea_t *)DRIGetSAREAPrivate(pScreen);
     if (pSAREAPriv->ctx_owner != (signed) DRIGetContext(pScreen)) {
-	struct R5xx3D *R5xx3D = rhdPtr->ThreeDPrivate;
+	if (rhdPtr->ChipSet < RHD_R600) {
+	    struct R5xx3D *R5xx3D = rhdPtr->ThreeDPrivate;
 
-	if (CS->Clean != RHD_CS_CLEAN_QUEUED) {
-	    R5xxDstCacheFlush(CS);
-	    R5xxZCacheFlush(CS);
-	    R5xxEngineWaitIdleFull(CS);
+	    if (CS->Clean != RHD_CS_CLEAN_QUEUED) {
+		R5xxDstCacheFlush(CS);
+		R5xxZCacheFlush(CS);
+		R5xxEngineWaitIdleFull(CS);
 
-	    CS->Clean = RHD_CS_CLEAN_QUEUED;
+		CS->Clean = RHD_CS_CLEAN_QUEUED;
+	    }
+
+	    if (R5xx3D)
+		R5xx3D->XHas3DEngineState = FALSE;
+	} else {
+	    struct r6xx_accel_state *accel_state = rhdPtr->TwoDPrivate;
+
+	    if (CS->Clean != RHD_CS_CLEAN_QUEUED) {
+		R6xxCacheFlush(CS);
+		R6xxEngineWaitIdleFull(CS);
+
+		CS->Clean = RHD_CS_CLEAN_QUEUED;
+	    }
+
+	    if (accel_state)
+		accel_state->XHas3DEngineState = FALSE;
 	}
-
-	if (R5xx3D)
-	    R5xx3D->XHas3DEngineState = FALSE;
     } else {
 	/* if the engine has been untouched, we need to track this too. */
 	if (CS->Clean != RHD_CS_CLEAN_QUEUED)
@@ -413,13 +431,18 @@ static void RHDEnterServer(ScreenPtr pScreen)
 static void RHDLeaveServer(ScreenPtr pScreen)
 {
     struct RhdCS *CS = RHDPTR(xf86Screens[pScreen->myNum])->CS;
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    RHDPtr rhdPtr = RHDPTR(pScrn);
 
     /* The CP is always running, but if we've generated any CP commands
      * we must flush them to the kernel module now. */
     if (CS->Clean == RHD_CS_CLEAN_DONE) {
-
-	R5xxDstCacheFlush(CS);
-	R5xxZCacheFlush(CS);
+	    if (rhdPtr->ChipSet >= RHD_R600) {
+		R6xxCacheFlush(CS);
+	    } else {
+		R5xxDstCacheFlush(CS);
+		R5xxZCacheFlush(CS);
+	    }
 
 	RHDCSFlush(CS); /* was a Release... */
 
@@ -504,51 +527,56 @@ static void RHDDRIInitGARTValues(struct rhdDri * rhdDRI)
 }
 
 /* Set AGP transfer mode according to requests and constraints */
-static Bool RHDSetAgpMode(struct rhdDri * rhdDRI, ScreenPtr pScreen)
+static Bool RHDSetAgpMode(struct rhdDri * rhdDRI)
 {
+    RHDPtr rhdPtr = RHDPTRI(rhdDRI);
     unsigned long mode   = drmAgpGetMode(rhdDRI->drmFD);	/* Default mode */
     unsigned int  vendor = drmAgpVendorId(rhdDRI->drmFD);
     unsigned int  device = drmAgpDeviceId(rhdDRI->drmFD);
-    /* ignore agp 3.0 mode bit from the chip as it's buggy on some cards with
-       pcie-agp rialto bridge chip - use the one from bridge which must match */
-    CARD32 agp_status = (RHDRegRead (rhdDRI, AGP_STATUS) | AGPv3_MODE) & mode;
-    Bool is_v3 = (agp_status & AGPv3_MODE);
 
-     RHDFUNC(rhdDRI);
+    if (rhdPtr->ChipSet < RHD_R600) {
+	/* ignore agp 3.0 mode bit from the chip as it's buggy on some cards with
+	   pcie-agp rialto bridge chip - use the one from bridge which must match */
+	CARD32 agp_status = (RHDRegRead (rhdDRI, AGP_STATUS) | AGPv3_MODE) & mode;
+	Bool is_v3 = (agp_status & AGPv3_MODE);
 
-   if (is_v3) {
-	rhdDRI->agpMode = (agp_status & AGPv3_8X_MODE) ? 8 : 4;
-    } else {
-	if (agp_status & AGP_4X_MODE)
-	    rhdDRI->agpMode = 4;
-	else if (agp_status & AGP_2X_MODE)
-	    rhdDRI->agpMode = 2;
-	else
-	    rhdDRI->agpMode = 1;
-    }
-    xf86DrvMsg(pScreen->myNum, X_DEFAULT, "Using AGP %dx\n", rhdDRI->agpMode);
+	RHDFUNC(rhdDRI);
 
-    mode &= ~AGP_MODE_MASK;
-    if (is_v3) {
-	/* only set one mode bit for AGPv3 */
-	switch (rhdDRI->agpMode) {
-	case 8:          mode |= AGPv3_8X_MODE; break;
-	case 4: default: mode |= AGPv3_4X_MODE;
+	if (is_v3) {
+	    rhdDRI->agpMode = (agp_status & AGPv3_8X_MODE) ? 8 : 4;
+	} else {
+	    if (agp_status & AGP_4X_MODE)
+		rhdDRI->agpMode = 4;
+	    else if (agp_status & AGP_2X_MODE)
+		rhdDRI->agpMode = 2;
+	    else
+		rhdDRI->agpMode = 1;
 	}
-    } else {
-	switch (rhdDRI->agpMode) {
-	case 4:          mode |= AGP_4X_MODE;
-	case 2:          mode |= AGP_2X_MODE;
-	case 1: default: mode |= AGP_1X_MODE;
-	}
-    }
+	xf86DrvMsg(rhdDRI->scrnIndex, X_DEFAULT, "Using AGP %dx\n", rhdDRI->agpMode);
 
-    xf86DrvMsg(pScreen->myNum, X_INFO,
+	mode &= ~AGP_MODE_MASK;
+	if (is_v3) {
+	    /* only set one mode bit for AGPv3 */
+	    switch (rhdDRI->agpMode) {
+	    case 8:          mode |= AGPv3_8X_MODE; break;
+	    case 4: default: mode |= AGPv3_4X_MODE;
+	    }
+	} else {
+	    switch (rhdDRI->agpMode) {
+	    case 4:          mode |= AGP_4X_MODE;
+	    case 2:          mode |= AGP_2X_MODE;
+	    case 1: default: mode |= AGP_1X_MODE;
+	    }
+	}
+    } else
+	rhdDRI->agpMode = 8; /* doesn't matter at this point */
+
+    xf86DrvMsg(rhdDRI->scrnIndex, X_INFO,
 	       "[agp] Mode 0x%08lx [AGP 0x%04x/0x%04x]\n",
 	       mode, vendor, device);
 
     if (drmAgpEnable(rhdDRI->drmFD, mode) < 0) {
-	xf86DrvMsg(pScreen->myNum, X_ERROR, "[agp] AGP not enabled\n");
+	xf86DrvMsg(rhdDRI->scrnIndex, X_ERROR, "[agp] AGP not enabled\n");
 	drmAgpRelease(rhdDRI->drmFD);
 	return FALSE;
     }
@@ -559,7 +587,10 @@ static Bool RHDSetAgpMode(struct rhdDri * rhdDRI, ScreenPtr pScreen)
 /* Initialize Radeon's AGP registers */
 static void RHDSetAgpBase(struct rhdDri * rhdDRI)
 {
-    RHDRegWrite (rhdDRI, AGP_BASE, drmAgpBase(rhdDRI->drmFD));
+    RHDPtr rhdPtr = RHDPTRI(rhdDRI);
+
+    if (rhdPtr->ChipSet < RHD_R600)
+	RHDRegWrite (rhdDRI, AGP_BASE, drmAgpBase(rhdDRI->drmFD));
 }
 
 /* Initialize the AGP state.  Request memory for use in AGP space, and
@@ -575,7 +606,7 @@ static Bool RHDDRIAgpInit(struct rhdDri * rhdDRI, ScreenPtr pScreen)
 	return FALSE;
     }
 
-    if (!RHDSetAgpMode(rhdDRI, pScreen))
+    if (!RHDSetAgpMode(rhdDRI))
 	return FALSE;
 
     RHDDRIInitGARTValues(rhdDRI);
@@ -850,17 +881,20 @@ static int RHDDRIKernelInit(RHDPtr rhdPtr, ScreenPtr pScreen)
     drmInfo.buffers_offset      = rhdDRI->bufHandle;
     drmInfo.gart_textures_offset= rhdDRI->gartTexHandle;
 
+    /* Make sure the MC has been set up before DRM_RADEON_CP_INIT is called */
+    ASSERT((RHD_CHECKDEBUGFLAG(rhdPtr, MC_SETUP)));
+
     if (drmCommandWrite(rhdDRI->drmFD, DRM_RADEON_CP_INIT,
 			&drmInfo, sizeof(drm_radeon_init_t)) < 0) {
 	xf86DrvMsg(pScreen->myNum, X_ERROR, "[dri] CP_INIT failed\n");
 	return FALSE;
     }
 
-    // FIXME: this is to be moved to rhd_cp
-    /* DRM_RADEON_CP_INIT does an engine reset, which resets some engine
-     * registers back to their default values, so we need to restore
-     * those engine register here. */
-//    R5xx2DSetup(pScrn);
+    /* - FIXME: this is to be moved to rhd_cp
+     * - DRM_RADEON_CP_INIT does an engine reset, which resets some engine
+     *   registers back to their default values, so we need to restore
+     *   those engine register here.
+     * - R5xx2DSetup(pScrn); */
 
     return TRUE;
 }
@@ -944,8 +978,8 @@ static void RHDDRIIrqInit(RHDPtr rhdPtr, ScreenPtr pScreen)
 		       "[drm] falling back to irq-free operation\n");
 	    rhdDRI->irq = 0;
 	} else {
-// FIXME
-//	    rhdDRI->ModeReg->gen_int_cntl = RHDRegRead (rhdDRI,  RADEON_GEN_INT_CNTL );
+/* FIXME 
+	    rhdDRI->ModeReg->gen_int_cntl = RHDRegRead (rhdDRI,  RADEON_GEN_INT_CNTL ); */
 	}
     }
 
@@ -1082,13 +1116,15 @@ static Bool RHDDRISetVBlankInterrupt(ScrnInfoPtr pScrn, Bool on)
 
     if (rhdDRI->irq) {
         if (on) {
-#ifdef RANDR_12_SUPPORT		// FIXME check / move to rhd_randr.c
-	    xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
-  	    if (xf86_config->num_crtc > 1 && xf86_config->crtc[1]->enabled)
-	        value = DRM_RADEON_VBLANK_CRTC1 | DRM_RADEON_VBLANK_CRTC2;
-	    else
+	    value = DRM_RADEON_VBLANK_CRTC1;
+
+#ifdef RANDR_12_SUPPORT		/* FIXME check / move to rhd_randr.c */
+	    if (RHDPTR(pScrn)->randr) {
+		xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
+		if (xf86_config->num_crtc > 1 && xf86_config->crtc[1]->enabled)
+		    value |= DRM_RADEON_VBLANK_CRTC2;
+	    }
 #endif
-	        value = DRM_RADEON_VBLANK_CRTC1;
 	}
 
 	if (RHDDRISetParam(pScrn, RADEON_SETPARAM_VBLANK_CRTC, value)) {
@@ -1109,9 +1145,11 @@ Bool RHDDRIPreInit(ScrnInfoPtr pScrn)
 
     RHDFUNC(rhdPtr);
 
+    rhdPtr->directRenderingEnabled = FALSE;
+
     if (!rhdPtr->useDRI.val.bool) {
-	xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Direct rendering turned off by"
-		   " default. Use Option \"DRI\" to enable.\n");
+	xf86DrvMsg(pScrn->scrnIndex, X_INFO,
+		   "Direct rendering explicitly turned off.\n");
 	return FALSE;
     }
 
@@ -1125,19 +1163,6 @@ Bool RHDDRIPreInit(ScrnInfoPtr pScrn)
 	return FALSE;
     }
 
-    if (rhdPtr->ChipSet >= RHD_R600) {
-	if (rhdPtr->useDRI.set && rhdPtr->useDRI.val.bool) {
-	    xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
-		       "Direct rendering for R600 an up forced on - "
-		       "This is NOT officially supported at the hardware level "
-		       "and may cause instability or lockups\n");
-	} else {
-	    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
-		       "Direct rendering not officially supported on R600 and up\n");
-	    return FALSE;
-	}
-    }
-
     if (!RHDDRIVersionCheck(rhdPtr)) {
 	xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
 		   "%s: Version check failed. Disabling DRI.\n", __func__);
@@ -1152,6 +1177,8 @@ Bool RHDDRIPreInit(ScrnInfoPtr pScrn)
     rhdDRI->ringSize      = RHD_DEFAULT_RING_SIZE;
     rhdDRI->bufSize       = RHD_DEFAULT_BUFFER_SIZE;
 
+    rhdDRI->drmFD         = -1;
+
     rhdDRI->gartLocation  = 0;
 
 #if 0
@@ -1244,7 +1271,6 @@ Bool RHDDRIAllocateBuffers(ScrnInfoPtr pScrn)
 	rhdPtr->FbFreeStart = old_freeoffset;
 	rhdPtr->FbFreeSize  = old_freesize;
 
-	/* return RHDDRICloseScreen(pScrn->pScreen); */
 	/* so far we are called from PreInit(): if we fail we free the DRI struct */
 	xfree(rhdPtr->dri);
 	rhdPtr->dri = NULL;
@@ -1347,12 +1373,15 @@ Bool RHDDRIScreenInit(ScreenPtr pScreen)
      * in the SAREA header */
     if (sizeof(XF86DRISAREARec)+sizeof(drm_radeon_sarea_t) > SAREA_MAX) {
 	ErrorF("Data does not fit in SAREA\n");
-	return RHDDRICloseScreen(pScreen);
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
     }
     pDRIInfo->SAREASize = SAREA_MAX;
 
-    if (!(pRADEONDRI = (RADEONDRIPtr)xcalloc(sizeof(RADEONDRIRec),1)))
-	return RHDDRICloseScreen(pScreen);
+    if (!(pRADEONDRI = (RADEONDRIPtr)xcalloc(sizeof(RADEONDRIRec),1))) {
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
+    }
 
     pDRIInfo->devPrivate     = pRADEONDRI;
     pDRIInfo->devPrivateSize = sizeof(RADEONDRIRec);
@@ -1375,7 +1404,8 @@ Bool RHDDRIScreenInit(ScreenPtr pScreen)
     if (!DRIScreenInit(pScreen, pDRIInfo, &rhdDRI->drmFD)) {
 	xf86DrvMsg(pScreen->myNum, X_ERROR,
 		   "[dri] DRIScreenInit failed.  Disabling DRI.\n");
-	return RHDDRICloseScreen(pScreen);
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
     }
 
     /* Initialize AGP */
@@ -1386,7 +1416,8 @@ Bool RHDDRIScreenInit(ScreenPtr pScreen)
 	xf86DrvMsg(pScreen->myNum, X_INFO,
 		   "[agp] You may want to make sure the agpgart kernel "
 		   "module\nis loaded before the radeon kernel module.\n");
-	return RHDDRICloseScreen(pScreen);
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
     }
 
     /* Initialize PCI */
@@ -1394,17 +1425,22 @@ Bool RHDDRIScreenInit(ScreenPtr pScreen)
 	!RHDDRIPciInit(rhdDRI, pScreen)) {
 	xf86DrvMsg(pScreen->myNum, X_ERROR,
 		   "[pci] PCI failed to initialize. Disabling the DRI.\n" );
-	return RHDDRICloseScreen(pScreen);
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
     }
 
     /* DRIScreenInit doesn't add all the common mappings.  Add additional
      * mappings here. */
-    if (!RHDDRIMapInit(rhdPtr, pScreen))
-	return RHDDRICloseScreen(pScreen);
+    if (!RHDDRIMapInit(rhdPtr, pScreen)) {
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
+    }
 
     /* FIXME: When are these mappings unmapped? */
-    if (!RHDInitVisualConfigs(pScreen))
-	return RHDDRICloseScreen(pScreen);
+    if (!RHDInitVisualConfigs(pScreen)) {
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
+    }
 
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[dri] Visual configs initialized\n");
 
@@ -1412,7 +1448,8 @@ Bool RHDDRIScreenInit(ScreenPtr pScreen)
     if (RHDDRISetParam(pScrn, RADEON_SETPARAM_NEW_MEMMAP, 1) < 0) {
 	xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
 		   "[drm] failed to enable new memory map\n");
-	return RHDDRICloseScreen(pScreen);
+	RHDDRICloseScreen(pScreen);
+	return FALSE;
     }
 
     return TRUE;
@@ -1428,12 +1465,6 @@ RHDDRIGARTBaseGet(RHDPtr rhdPtr)
     drm_radeon_getparam_t gp;
     int gart_base;
 
-    if (rhdPtr->cardType == RHD_CARD_AGP) {
-	xf86DrvMsg(rhdPtr->scrnIndex, X_INFO,
-		   "%s: Unable to get GART address (AGP card).\n", __func__);
-	return 0;
-    }
-
     memset(&gp, 0, sizeof(gp));
     gp.param = RADEON_PARAM_GART_BASE;
     gp.value = &gart_base;
@@ -1444,7 +1475,7 @@ RHDDRIGARTBaseGet(RHDPtr rhdPtr)
 		   "%s: Failed to determine GART area MC location.\n", __func__);
 	return 0;
     } else {
-	RHDDebug(rhdPtr->scrnIndex, "GART location: 0x08X\n", gart_base);
+	RHDDebug(rhdPtr->scrnIndex, "GART location: 0x%08X\n", gart_base);
 	return gart_base;
     }
 }
@@ -1576,16 +1607,15 @@ void RHDDRIEnterVT(ScreenPtr pScreen)
 
     RHDFUNC(rhdPtr);
 
+    if (rhdDRI->drmFD == -1)
+	return;
+
     if (rhdPtr->cardType == RHD_CARD_AGP) {
-	if (!RHDSetAgpMode(rhdDRI, pScreen))
+	if (!RHDSetAgpMode(rhdDRI))
 	    return;
 	RHDSetAgpBase(rhdDRI);
     }
 
-    if ( (ret = drmCommandNone(rhdDRI->drmFD, DRM_RADEON_CP_RESUME)) )
-	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
-		   "%s: CP resume %d\n", __func__, ret);
-
     /* TODO: maybe using CP_INIT instead of CP_RESUME is enough, so we wouldn't
      * need an additional copy of the GART table in main memory. OTOH the table
      * must be initialized but not allocated anew. */
@@ -1596,7 +1626,9 @@ void RHDDRIEnterVT(ScreenPtr pScreen)
 
     RHDDRISetVBlankInterrupt(pScrn, rhdDRI->have3Dwindows);
 
-    DRIUnlock(pScrn->pScreen);
+    if ( (ret = drmCommandNone(rhdDRI->drmFD, DRM_RADEON_CP_RESUME)) )
+	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+		   "%s: CP resume %d\n", __func__, ret);
 }
 
 /* Stop all before vt switch / suspend */
@@ -1608,6 +1640,9 @@ void RHDDRILeaveVT(ScreenPtr pScreen)
 
     RHDFUNC(rhdPtr);
 
+    if (rhdDRI->drmFD == -1)
+	return;
+
     RHDDRISetVBlankInterrupt (pScrn, FALSE);
     DRILock(pScrn->pScreen, 0);
 
@@ -1636,7 +1671,7 @@ Bool RHDDRICloseScreen(ScreenPtr pScreen)
     ScrnInfoPtr    pScrn  = xf86Screens[pScreen->myNum];
     RHDPtr         rhdPtr = RHDPTR(pScrn);
     struct rhdDri *rhdDRI   = rhdPtr->dri;
-    //drm_radeon_init_t drmInfo;
+    /* drm_radeon_init_t drmInfo; */
     drm_radeon_init_t drmInfo;
 
     RHDFUNC(pScrn);
@@ -1645,7 +1680,7 @@ Bool RHDDRICloseScreen(ScreenPtr pScreen)
 	RHDDRISetVBlankInterrupt (pScrn, FALSE);
 	drmCtlUninstHandler(rhdDRI->drmFD);
 	rhdDRI->irq = 0;
-//	rhdDRI->ModeReg->gen_int_cntl = 0;
+/*	rhdDRI->ModeReg->gen_int_cntl = 0; */
     }
 
     /* invalidate GART location for EXA */
@@ -1697,6 +1732,7 @@ Bool RHDDRICloseScreen(ScreenPtr pScreen)
 
     /* De-allocate all DRI resources */
     DRICloseScreen(pScreen);
+    rhdDRI->drmFD = -1;
 
     /* De-allocate all DRI data structures */
     if (rhdDRI->pDRIInfo) {
@@ -1716,10 +1752,9 @@ Bool RHDDRICloseScreen(ScreenPtr pScreen)
 	rhdDRI->pVisualConfigsPriv = NULL;
     }
 
-    xfree(rhdDRI);
-    rhdPtr->dri = NULL;
+    rhdPtr->directRenderingEnabled = FALSE;
 
-    return FALSE;
+    return TRUE;
 }
 
 
@@ -1736,8 +1771,8 @@ static void RHDDRITransitionSingleToMulti3d(ScreenPtr pScreen)
 
 static void RHDDRITransitionMultiToSingle3d(ScreenPtr pScreen)
 {
-    /* Let the remaining 3d app start page flipping again */
-//    RHDEnablePageFlip(pScreen);
+    /* Let the remaining 3d app start page flipping again 
+     * RHDEnablePageFlip(pScreen); */
 }
 
 static void RHDDRITransitionTo3d(ScreenPtr pScreen)
@@ -1746,8 +1781,8 @@ static void RHDDRITransitionTo3d(ScreenPtr pScreen)
     struct rhdDri *rhdDRI   = RHDPTR(pScrn)->dri;
 
     rhdDRI->have3Dwindows = TRUE;
-//    RHDChangeSurfaces(pScrn);	// FIXME needed for tiling
-//    RHDEnablePageFlip(pScreen);
+/*    RHDChangeSurfaces(pScrn);	*//* FIXME needed for tiling */
+/*    RHDEnablePageFlip(pScreen); */
 
     RHDDRISetVBlankInterrupt(pScrn, TRUE);
 }
@@ -1772,20 +1807,25 @@ static void RHDDRITransitionTo2d(ScreenPtr pScreen)
 		   "[dri] RHDDRITransitionTo2d: "
 		   "kernel failed to unflip buffers.\n");
     }
-//    RHDChangeSurfaces(pScrn);
+/*   RHDChangeSurfaces(pScrn); */
 
     RHDDRISetVBlankInterrupt(pScrn, FALSE);
 }
 
 static int RHDDRIGetPciAperTableSize(ScrnInfoPtr pScrn)
 {
+    RHDPtr rhdPtr = RHDPTR(pScrn);
     int page_size  = getpagesize();
     int ret_size;
     int num_pages;
 
     num_pages = (RHD_DEFAULT_PCI_APER_SIZE * 1024 * 1024) / page_size;
 
-    ret_size = num_pages * sizeof(unsigned int);
+    if ((rhdPtr->ChipSet >= RHD_R600) ||
+	(rhdPtr->ChipSet == RHD_RS600))
+	ret_size = num_pages * sizeof(uint64_t);
+    else
+	ret_size = num_pages * sizeof(unsigned int);
 
     return ret_size;
 }
@@ -1874,7 +1914,8 @@ RHDDRMCPBuffer(int scrnIndex)
 	int ret = drmDMA(Dri->drmFD, &dma);
 	if (!ret) {
 	    buf = &Dri->buffers->list[indx];
-	    /* RHDDebug(scrnIndex, "%s: index %d, addr %p\n",  __func__, buf->idx, buf->address); */
+	    /* xf86DrvMsg(scrnIndex, X_INFO, "%s: index %d, addr %p\n",  __func__, buf->idx, buf->address); */
+	    buf->used = 0;
 	    return buf;
 	} else if (ret != -16)
 	    xf86DrvMsg(scrnIndex, X_ERROR, "%s: drmDMA returned %d\n",
@@ -1940,3 +1981,15 @@ RHDDRMIndirectBufferDiscard(int scrnIndex, CARD8 *Buffer)
 	       "%s: Unable to retrieve the indirect Buffer at address %p!\n",
 	       __func__, Buffer);
 }
+
+unsigned int
+RHDDRIGetIntGARTLocation(ScrnInfoPtr pScrn)
+{
+    RHDPtr         rhdPtr  = RHDPTR(pScrn);
+    struct rhdDri *rhdDRI    = rhdPtr->dri;
+
+    if (!rhdDRI->gartLocation)
+        return 0;
+
+    return rhdDRI->gartLocation + rhdDRI->bufStart;
+}
